unit MCoPRecentMenu;

{------------------------------------}
{         TRecentMenu 1.0.1          }
{        By Mohammad Yousefi         }
{              2013                  }
{      vahid_you2004@yahoo.com       }
{       http://www.dihav.com/        }
{------------------------------------}

interface

uses
  SysUtils, Classes,Registry, Windows, ShlObj, ShellApi, CommCtrl,
  {$if CompilerVersion > 22}
    Vcl.Menus, Vcl.ImgList, Vcl.Graphics
  {$else}
    Menus, ImgList, Graphics
  {$ifend}
  ;

type

  TCaptionFormat = (cfFileName, cfFileNameWithExtension, cfFileAddress);

  TRecentItem = class(TMenuItem)
  private
    fIndex:Integer;
  protected
    property Index:Integer read fIndex write fIndex;
  end;

  PMenuItemList = ^TMenuItemList;
  TMenuItemList = record
   Prv,Nxt:PMenuItemList;
   Mn:TRecentItem;
   Index:Integer;
  end;

  TRecentMenuParentItems = class;

  TRecentMenuParentItem = class(TCollectionItem)
  private
    fMaxWidth:Cardinal;
    fInsertionIndex,fImageIndex,fMaxNumberOfFiles:Integer;
    fHint:String;
    fExtractFileIconFromShell,fDisplayFileNumber:Boolean;
    fParentItem:TMenuItem;
    fItemsList:PMenuItemList;
    fCaptionFormat:TCaptionFormat;
    fParentMenu:TPopupMenu;
    function GetItem(Index:Integer):TMenuItem;
    procedure IClk(Sender:TObject);
    function IsObject(Obj:TObject):Boolean;
    procedure SetParentItem(V:TMenuItem);
    procedure SetParentMenu(V:TPopupMenu);
  protected
    function GetDisplayName:String; override;
    procedure RefreshItems;
    procedure RemoveMenu(AComponent:TComponent);
  public
    constructor Create(Collection:TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source:TPersistent); override;
    property Items[Index:Integer]:TMenuItem read GetItem;
  published
    property MaxWidth:Cardinal read fMaxWidth write fMaxWidth default 0;
    property ImageIndex:Integer read fImageIndex write fImageIndex default -1;
    property Hint:String read fHint write fHint;
    property ExtractFileIconFromShell:Boolean read fExtractFileIconFromShell write fExtractFileIconFromShell default True;
    property ParentItem:TMenuItem read fParentItem write SetParentItem;
    property InsertionIndex:Integer read fInsertionIndex write fInsertionIndex default 0;
    property DisplayFileNumber:Boolean read fDisplayFileNumber write fDisplayFileNumber default True;
    property CaptionFormat:TCaptionFormat read fCaptionFormat write fCaptionFormat default cfFileAddress;
    property ParentMenu:TPopupMenu read fParentMenu write SetParentMenu;
    property MaxNumberOfFiles:Integer read fMaxNumberOfFiles write fMaxNumberOfFiles default 4;
  end;

  TRecentMenu = class;

  TRecentMenuParentItems = class(TOwnedCollection)
  private
    function GetItem(Index:Integer):TRecentMenuParentItem;
    procedure SetItem(Index:Integer;Value:TRecentMenuParentItem);
  public
    constructor Create(AOwner:TPersistent);
    function Add:TRecentMenuParentItem;
    function Insert(Index:Integer):TRecentMenuParentItem;
    property Items[Index:Integer]:TRecentMenuParentItem read GetItem write SetItem; default;
  end;

  TRecentItemClick = procedure (Sender:TMenuItem;FileName:TFileName) of object;
  TRecentItemChange = procedure (Sender:TRecentMenu;Item:TMenuItem;FileName:TFileName) of object;

  TRecentMenu = class(TComponent)
  private
    fRegistryKey:String;
    fMaxNumberOfFiles:Integer;
    fAddToWindowsRecent:Boolean;
    fMenus:TRecentMenuParentItems;
    fEvent:TRecentItemClick;
    fCEvent,fDEvent:TRecentItemChange;
    procedure SetHP(V:String);
    function GetHP:String;
  protected
    ListOfFiles:TStringList;
    procedure ItemClick(Sender:TRecentItem;Index:Integer);
    procedure ItemCreate(Sender:TRecentItem;Index:Integer);
    procedure ItemDestroy(Sender:TRecentItem;Index:Integer);
    procedure Notification(AComponent:TComponent;Operation:TOperation); override;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure AddFile(FileName:TFileName);
    procedure MakeRecentMenus;
  published
    property OnClick:TRecentItemClick read fEvent write fEvent;
    property OnCreateItem:TRecentItemChange read fCEvent write fCEvent;
    property OnDestroyItem:TRecentItemChange read fDEvent write fDEvent;
    property RegistryKey:String read fRegistryKey write fRegistryKey;
    property MaxNumberOfFiles:Integer read fMaxNumberOfFiles write fMaxNumberOfFiles default 4;
    property AddToWindowsRecent:Boolean read fAddToWindowsRecent write fAddToWindowsRecent default True;
    property Menus:TRecentMenuParentItems read fMenus write fMenus;
    property Homepage:String read GetHP write SetHP stored False;
  end;

procedure Register;

implementation

procedure TRecentMenuParentItem.Assign(Source:TPersistent);
begin
if Source is TRecentMenuParentItem then
 begin
 fMaxWidth:=TRecentMenuParentItem(Source).MaxWidth;
 fImageIndex:=TRecentMenuParentItem(Source).ImageIndex;
 fInsertionIndex:=TRecentMenuParentItem(Source).InsertionIndex;
 fHint:=TRecentMenuParentItem(Source).Hint;
 fExtractFileIconFromShell:=TRecentMenuParentItem(Source).ExtractFileIconFromShell;
 fParentItem:=TRecentMenuParentItem(Source).ParentItem;
 fDisplayFileNumber:=TRecentMenuParentItem(Source).DisplayFileNumber;
 fCaptionFormat:=TRecentMenuParentItem(Source).CaptionFormat;
 end
else
 inherited Assign(Source);
end;

constructor TRecentMenuParentItem.Create(Collection:TCollection);
begin
fMaxWidth:=0;
fImageIndex:=-1;
fInsertionIndex:=0;
fHint:='Opens %s';
fExtractFileIconFromShell:=True;
fDisplayFileNumber:=True;
fParentItem:=nil;
fItemsList:=nil;
fMaxNumberOfFiles:=4;
fCaptionFormat:=cfFileAddress;
inherited Create(Collection);
end;

destructor TRecentMenuParentItem.Destroy;
begin
if fItemsList<>nil then
 begin
 while fItemsList^.Prv<>nil do
  begin
  fItemsList:=fItemsList^.Prv;
  FreeMem(fItemsList^.Nxt);
  end;
 FreeMem(fItemsList);
 end;
inherited Destroy;
end;

function TRecentMenuParentItem.GetDisplayName:String;
var s:string;
begin
s:='<unassigned menu>';
if IsObject(fParentItem) then s:=fParentItem.Name;
if IsObject(fParentMenu) then s:=fParentMenu.Name;
Result:=s+' recent items';
end;

function TRecentMenuParentItem.GetItem(Index:Integer):TMenuItem;
var i:PMenuItemList;
begin
if fItemsList=nil then
 begin
 Result:=nil;
 exit;
 end;
i:=fItemsList;
if i^.Index=Index then Result:=i^.Mn else Result:=nil;
if Result=nil then
 while i^.Prv<>nil do
  begin
  i:=i^.Prv;
  if i^.Index=Index then
   begin
   Result:=i^.Mn;
   break;
   end;
  end;
end;

procedure TRecentMenuParentItem.RefreshItems;
var
 k,a,x,y,u:Integer;
 m:TRecentItem;
 s:String;
 z:TSize;
 d:HDC;
 l:TStringList;
 p:TMenuItem;
 f:TSHFILEINFO;
 b,c,e:{$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap;
begin
p:=nil;
if IsObject(fParentItem) then p:=fParentItem;
if IsObject(fParentMenu) then p:=fParentMenu.Items;
if p=nil then exit;

if fItemsList<>nil then
 begin
 while fItemsList^.Prv<>nil do
  begin
  ((GetOwner as TRecentMenuParentItems).GetOwner as TRecentMenu).ItemDestroy(fItemsList^.Mn,fItemsList^.Mn.Index);
  fItemsList^.Mn.Destroy;
  fItemsList:=fItemsList^.Prv;
  FreeMem(fItemsList^.Nxt);
  end;
 fItemsList^.Mn.Destroy;
 FreeMem(fItemsList);
 fItemsList:=nil;
 end;

l:=((GetOwner as TRecentMenuParentItems).GetOwner as TRecentMenu).ListOfFiles;
u:=fMaxNumberOfFiles;
if u>l.Count then u:=l.Count;
if u<>0 then
 for k:=u-1 downto 0 do
  begin
  m:=TRecentItem.Create(fParentItem);
  m.Index:=k;
  s:=l.Strings[k];
  if fCaptionFormat<>cfFileAddress then s:=extractfilename(s);
  if fCaptionFormat=cfFileName then s:=changefileext(s,'');
  d:=GetDC(HWND_DESKTOP);
  GetTextExtentPoint32(d,PChar(s),Length(s),z);
  if (fMaxWidth<Cardinal(z.cx)) and (fMaxWidth>0) then
   begin
   insert('...',s,4);
   while Cardinal(z.cx)>fMaxWidth do
    begin
    delete(s,7,1);
    GetTextExtentPoint32(d,PChar(s),Length(s),z);
    end;
   end;
  if fDisplayFileNumber then s:=format('&%d %s',[k+1,s]);
  m.Caption:=s;
  m.Hint:=format(fHint,[l.Strings[k]]);
  m.OnClick:=IClk;
  m.ImageIndex:=fImageIndex;
  if fExtractFileIconFromShell then
   if p.GetImageList<>nil then
    begin
    a:=SHGFI_SYSICONINDEX;
    if p.GetImageList.Width>16 then a:=a or SHGFI_LARGEICON else a:=a or SHGFI_SMALLICON;
    a:=shgetfileinfo(pchar(l.Strings[k]),0,f,sizeof(f),a);
    ImageList_GetIconSize(a,x,y);
    e:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
    e.Width:=x;
    e.Height:=y;
    ImageList_Draw(a,f.iIcon,e.Canvas.Handle,0,0,ILD_NORMAL);
    b:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
    b.Width:=p.GetImageList.Width;
    b.Height:=p.GetImageList.Height;
    b.Canvas.CopyRect(b.Canvas.ClipRect,e.Canvas,e.Canvas.ClipRect);
    e.Free;
    e:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
    e.Width:=x;
    e.Height:=y;
    e.Monochrome:=True;
    ImageList_Draw(a,f.iIcon,e.Canvas.Handle,0,0,ILD_MASK);
    c:={$if CompilerVersion > 22}Vcl.{$ifend}Graphics.TBitmap.Create;
    c.Width:=b.Width;
    c.Height:=b.Height;
    c.Monochrome:=True;
    c.Canvas.CopyRect(c.Canvas.ClipRect,e.Canvas,e.Canvas.ClipRect);
    e.Free;
    m.ImageIndex:=p.GetImageList.Add(b,c);
    b.Free;
    c.Free;
    end;
  p.Insert(fInsertionIndex,m);
  if fItemsList=nil then
   begin
   fItemsList:=AllocMem(SizeOf(TMenuItemList));
   fItemsList^.Prv:=nil;
   fItemsList^.Nxt:=nil;
   fItemsList^.Index:=k;
   end
  else
   begin
   fItemsList^.Nxt:=AllocMem(SizeOf(TMenuItemList));
   fItemsList^.Nxt^.Prv:=fItemsList;
   fItemsList:=fItemsList^.Nxt;
   fItemsList^.Nxt:=nil;
   fItemsList^.Index:=k;
   end;
  fItemsList.Mn:=m;
  ((GetOwner as TRecentMenuParentItems).GetOwner as TRecentMenu).ItemCreate(m,m.Index);
  end;
end;

procedure TRecentMenuParentItem.IClk(Sender:TObject);
var m:TRecentItem;
begin
m:=Sender as TRecentItem;
((GetOwner as TRecentMenuParentItems).GetOwner as TRecentMenu).ItemClick(m,m.Index);
end;

function TRecentMenuParentItem.IsObject(Obj:TObject):Boolean;
begin
Result:=Obj<>nil;
end;

procedure TRecentMenuParentItem.RemoveMenu(AComponent:TComponent);
begin
if AComponent=fParentItem then fParentItem:=nil;
if AComponent=fParentMenu then fParentMenu:=nil;
end;

procedure TRecentMenuParentItem.SetParentItem(V:TMenuItem);
begin
if V<>nil then fParentMenu:=nil;
fParentItem:=V;
end;

procedure TRecentMenuParentItem.SetParentMenu(V:TPopupMenu);
begin
if V<>nil then fParentItem:=nil;
fParentMenu:=V;
end;



function TRecentMenuParentItems.Add:TRecentMenuParentItem;
begin
Result:=TRecentMenuParentItem(inherited Add);
end;

function TRecentMenuParentItems.GetItem(Index:Integer):TRecentMenuParentItem;
begin
Result:=TRecentMenuParentItem(inherited GetItem(Index));
end;

procedure TRecentMenuParentItems.SetItem(Index:Integer;Value:TRecentMenuParentItem);
begin
inherited SetItem(Index,Value);
end;

function TRecentMenuParentItems.Insert(Index:Integer):TRecentMenuParentItem;
begin
Result:=TRecentMenuParentItem(inherited Insert(Index));
end;

constructor TRecentMenuParentItems.Create(AOwner:TPersistent);
begin
inherited Create(AOwner,TRecentMenuParentItem);
end;



constructor TRecentMenu.Create(AOwner:TComponent);
begin
fRegistryKey:='Software\Your Company\Your Product\RecentFiles';
fMaxNumberOfFiles:=4;
fAddToWindowsRecent:=True;
inherited Create(AOwner);
fMenus:=TRecentMenuParentItems.Create(Self);
ListOfFiles:=TStringList.Create;
ListOfFiles.CaseSensitive:=False;
end;

destructor TRecentMenu.Destroy;
begin
fMenus.Free;
ListOfFiles.Free;
inherited Destroy;
end;

procedure TRecentMenu.ItemClick(Sender:TRecentItem;Index:Integer);
begin
if Assigned(fEvent) then
 fEvent(Sender,ListOfFiles.Strings[Index]);
end;

procedure TRecentMenu.ItemCreate(Sender:TRecentItem;Index:Integer);
begin
if Assigned(fCEvent) then
 fCEvent(Self,Sender,ListOfFiles.Strings[Index]);
end;

procedure TRecentMenu.ItemDestroy(Sender:TRecentItem;Index:Integer);
begin
if Assigned(fDEvent) then
 fDEvent(Self,Sender,ListOfFiles.Strings[Index]);
end;

procedure TRecentMenu.AddFile(FileName:TFileName);
var
 k:Integer;
 r:TRegistry;
begin
k:=ListOfFiles.IndexOf(FileName);
if k>=0 then
 ListOfFiles.Move(k,0)
else
 ListOfFiles.Insert(0,FileName);
while ListOfFiles.Count>fMaxNumberOfFiles do
 ListOfFiles.Delete(ListOfFiles.Count-1);
if fAddToWindowsRecent then SHAddToRecentDocs(SHARD_PATH,@FileName);
for k:=0 to fMenus.Count-1 do
 fMenus.Items[k].RefreshItems;
r:=TRegistry.Create;
if r.OpenKey(fRegistryKey,True) then
 for k:=0 to ListOfFiles.Count-1 do
  r.WriteString('RecentFile'+inttostr(k),ListOfFiles.Strings[k]);
r.Free;
end;

procedure TRecentMenu.MakeRecentMenus;
var
 k:Integer;
 r:TRegistry;
begin
r:=TRegistry.Create;
k:=0;
if r.OpenKeyReadOnly(fRegistryKey) then
 while r.ValueExists('RecentFile'+inttostr(k)) do
  begin
  ListOfFiles.Add(r.ReadString('RecentFile'+inttostr(k)));
  inc(k);
  end;
r.Free;
while ListOfFiles.Count>fMaxNumberOfFiles do
 ListOfFiles.Delete(ListOfFiles.Count-1);
for k:=0 to fMenus.Count-1 do
 fMenus.Items[k].RefreshItems;
end;

procedure TRecentMenu.SetHP(V:String);
begin
//
end;

function TRecentMenu.GetHP:String;
begin
Result:='http://www.dihav.com/';
end;

procedure TRecentMenu.Notification(AComponent:TComponent;Operation:TOperation);
var k:Integer;
begin
inherited Notification(AComponent,Operation);
if (Operation=opRemove) and ((AComponent is TPopupMenu) or (AComponent is TMenuItem)) then
 for k:=0 to fMenus.Count-1 do
  fMenus.Items[k].RemoveMenu(AComponent);
end;



procedure Register;
begin
  RegisterComponents('dihav', [TRecentMenu]);
end;

end.
 